@@ -44,6 +44,12 @@
 #define MAX_DELEGATE	50
 #define MAX_CATAL_DEPTH	50
 
+#ifdef _WIN32
+# define PATH_SEAPARATOR ';'
+#else
+# define PATH_SEAPARATOR ':'
+#endif
+
 /**
  * TODO:
  *
@@ -71,9 +77,15 @@
 #if defined(_WIN32) && defined(_MSC_VER)
 #undef XML_XML_DEFAULT_CATALOG
 static char XML_XML_DEFAULT_CATALOG[256] = "file:///etc/xml/catalog";
+#if defined(_WIN32_WCE)
+/* Windows CE don't have a A variant */
+#define GetModuleHandleA GetModuleHandle
+#define GetModuleFileNameA GetModuleFileName
+#else
 void* __stdcall GetModuleHandleA(const char*);
 unsigned long __stdcall GetModuleFileNameA(void*, char*, unsigned long);
 #endif
+#endif
 
 static xmlChar *xmlCatalogNormalizePublic(const xmlChar *pubID);
 static int xmlExpandCatalog(xmlCatalogPtr catal, const char *filename);
@@ -1200,8 +1212,6 @@ static void
 xmlParseXMLCatalogNode(xmlNodePtr cur, xmlCatalogPrefer prefer,
 	               xmlCatalogEntryPtr parent, xmlCatalogEntryPtr cgroup)
 {
-    xmlChar *uri = NULL;
-    xmlChar *URL = NULL;
     xmlChar *base = NULL;
     xmlCatalogEntryPtr entry = NULL;
 
@@ -1288,10 +1298,6 @@ xmlParseXMLCatalogNode(xmlNodePtr cur, xmlCatalogPrefer prefer,
     }
     if (base != NULL)
 	xmlFree(base);
-    if (uri != NULL)
-	xmlFree(uri);
-    if (URL != NULL)
-	xmlFree(URL);
 }
 
 /**
@@ -1546,6 +1552,7 @@ xmlAddXMLCatalog(xmlCatalogEntryPtr catal, const xmlChar *type,
 	cur->next = xmlNewCatalogEntry(typ, orig, replace,
 		                       NULL, catal->prefer, NULL);
     if (doregister) {
+        catal->type = XML_CATA_CATALOG;
 	cur = xmlHashLookup(xmlCatalogXMLFiles, catal->URL);
 	if (cur != NULL)
 	    cur->children = catal->children;
@@ -1648,7 +1655,8 @@ xmlCatalogXMLResolve(xmlCatalogEntryPtr catal, const xmlChar *pubID,
 		    if (xmlStrEqual(sysID, cur->name)) {
 			if (xmlDebugCatalogs)
 			    xmlGenericError(xmlGenericErrorContext,
-				    "Found system match %s\n", cur->name);
+				    "Found system match %s, using %s\n",
+				            cur->name, cur->URL);
 			catal->depth--;
 			return(xmlStrdup(cur->URL));
 		    }
@@ -3213,24 +3221,35 @@ xmlLoadCatalogs(const char *pathss) {
     const char *cur;
     const char *paths;
     xmlChar *path;
+#ifdef _WIN32
+    int i, iLen;
+#endif
 
     if (pathss == NULL)
 	return;
 
     cur = pathss;
-    while ((cur != NULL) && (*cur != 0)) {
+    while (*cur != 0) {
 	while (xmlIsBlank_ch(*cur)) cur++;
 	if (*cur != 0) {
 	    paths = cur;
-	    while ((*cur != 0) && (*cur != ':') && (!xmlIsBlank_ch(*cur)))
+	    while ((*cur != 0) && (*cur != PATH_SEAPARATOR) && (!xmlIsBlank_ch(*cur)))
 		cur++;
 	    path = xmlStrndup((const xmlChar *)paths, cur - paths);
+#ifdef _WIN32
+        iLen = strlen(path);
+        for(i = 0; i < iLen; i++) {
+            if(path[i] == '\\') {
+                path[i] = '/';
+            }
+        }
+#endif
 	    if (path != NULL) {
 		xmlLoadCatalog((const char *) path);
 		xmlFree(path);
 	    }
 	}
-	while (*cur == ':')
+	while (*cur == PATH_SEAPARATOR)
 	    cur++;
     }
 }
