@@ -2343,9 +2343,13 @@ done:
 	 (((p[0] >= 'a') && (p[0] <= 'z')) ||			\
 	  ((p[0] >= 'A') && (p[0] <= 'Z'))) &&			\
 	 (p[1] == ':') && ((p[2] == '/') || (p[2] == '\\')))
-xmlChar*
+xmlChar *
 xmlCanonicPath(const xmlChar *path)
 {
+/*
+ * For Windows implementations, additional work needs to be done to
+ * replace backslashes in pathnames with "forward slashes"
+ */
 #if defined(_WIN32) && !defined(__CYGWIN__)    
     int len = 0;
     int i = 0;
@@ -2362,6 +2366,7 @@ xmlCanonicPath(const xmlChar *path)
 	return xmlStrdup(path);
     }
 
+    /* Check if this is an "absolute uri" */
     absuri = xmlStrstr(path, BAD_CAST "://");
     if (absuri != NULL) {
         int l, j;
@@ -2370,65 +2375,78 @@ xmlCanonicPath(const xmlChar *path)
 
         /*
 	 * this looks like an URI where some parts have not been
-	 * escaped leading to a parsing problem check that the first
+	 * escaped leading to a parsing problem.  Check that the first
 	 * part matches a protocol.
 	 */
 	l = absuri - path;
+	/* Bypass if first part (part before the '://') is > 20 chars */
 	if ((l <= 0) || (l > 20))
 	    goto path_processing;
+	/* Bypass if any non-alpha characters are present in first part */
 	for (j = 0;j < l;j++) {
 	    c = path[j];
 	    if (!(((c >= 'a') && (c <= 'z')) || ((c >= 'A') && (c <= 'Z'))))
 	        goto path_processing;
 	}
 
+	/* Escape all except the characters specified in the supplied path */
         escURI = xmlURIEscapeStr(path, BAD_CAST ":/?_.#&;=");
 	if (escURI != NULL) {
+	    /* Try parsing the escaped path */
 	    uri = xmlParseURI((const char *) escURI);
+	    /* If successful, return the escaped string */
 	    if (uri != NULL) {
 	        xmlFreeURI(uri);
 		return escURI;
 	    }
-	    xmlFreeURI(uri);
 	}
     }
 
 path_processing:
+/* For Windows implementations, replace backslashes with 'forward slashes' */
 #if defined(_WIN32) && !defined(__CYGWIN__)    
     /*
-     * This really need to be cleaned up by someone with a Windows box
+     * Create a URI structure
      */
     uri = xmlCreateURI();
-    if (uri == NULL) {
+    if (uri == NULL) {		/* Guard against 'out of memory' */
         return(NULL);
     }
 
     len = xmlStrlen(path);
     if ((len > 2) && IS_WINDOWS_PATH(path)) {
+        /* make the scheme 'file' */
 	uri->scheme = xmlStrdup(BAD_CAST "file");
-	uri->path = xmlMallocAtomic(len + 2);	/* FIXME - check alloc! */
+	/* allocate space for leading '/' + path + string terminator */
+	uri->path = xmlMallocAtomic(len + 2);
+	if (uri->path == NULL) {
+	    xmlFreeURI(uri);	/* Guard agains 'out of memory' */
+	    return(NULL);
+	}
+	/* Put in leading '/' plus path */
 	uri->path[0] = '/';
 	p = uri->path + 1;
 	strncpy(p, path, len + 1);
     } else {
-	uri->path = xmlStrdup(path);		/* FIXME - check alloc! */
+	uri->path = xmlStrdup(path);
+	if (uri->path == NULL) {
+	    xmlFreeURI(uri);
+	    return(NULL);
+	}
 	p = uri->path;
     }
+    /* Now change all occurences of '\' to '/' */
     while (*p != '\0') {
 	if (*p == '\\')
 	    *p = '/';
 	p++;
     }
-    if (uri->path == NULL) {
-        xmlFreeURI(uri);
-        return(NULL);
-    }
 
-	if (uri->scheme == NULL) {
-		ret = xmlStrdup((const xmlChar *) path);
-	} else {
-		ret = xmlSaveUri(uri);
-	}
+    if (uri->scheme == NULL) {
+	ret = xmlStrdup((const xmlChar *) uri->path);
+    } else {
+	ret = xmlSaveUri(uri);
+    }
 
     xmlFreeURI(uri);
 #else
@@ -2437,5 +2455,57 @@ path_processing:
     return(ret);
 }
 
+/**
+ * xmlPathToURI:
+ * @path:  the resource locator in a filesystem notation
+ *
+ * Constructs an URI expressing the existing path
+ *
+ * Returns a new URI, or a duplicate of the path parameter if the 
+ * construction fails. The caller is responsible for freeing the memory
+ * occupied by the returned string. If there is insufficient memory available,
+ * or the argument is NULL, the function returns NULL.
+ */
+xmlChar *
+xmlPathToURI(const xmlChar *path)
+{
+    xmlURIPtr uri;
+    xmlURI temp;
+    xmlChar *ret, *cal;
+
+    if (path == NULL)
+        return(NULL);
+
+    if ((uri = xmlParseURI((const char *) path)) != NULL) {
+	xmlFreeURI(uri);
+	return xmlStrdup(path);
+    }
+    cal = xmlCanonicPath(path);
+    if (cal == NULL)
+        return(NULL);
+#if defined(_WIN32) && !defined(__CYGWIN__)
+    /* xmlCanonicPath can return an URI on Windows (is that the intended behaviour?) 
+       If 'cal' is a valid URI allready then we are done here, as continuing would make
+       it invalid. */
+    if ((uri = xmlParseURI((const char *) cal)) != NULL) {
+	xmlFreeURI(uri);
+	return cal;
+    }
+    /* 'cal' can contain a relative path with backslashes. If that is processed
+       by xmlSaveURI, they will be escaped and the external entity loader machinery
+       will fail. So convert them to slashes. Misuse 'ret' for walking. */
+    ret = cal;
+    while (*ret != '\0') {
+	if (*ret == '\\')
+	    *ret = '/';
+	ret++;
+    }
+#endif
+    memset(&temp, 0, sizeof(temp));
+    temp.path = (char *) cal;
+    ret = xmlSaveUri(&temp);
+    xmlFree(cal);
+    return(ret);
+}
 #define bottom_uri
 #include "elfgcchack.h"
